use log::{LevelFilter, Log};
use serde::{Deserialize, Serialize};
use std::{collections, error::Error};

use tarantool::{
    log::TarantoolLogger,
    proc,
    space::{self},
    tlua, tuple,
};

const SPACE_SPACE_ID: u32 = 280;
const INDEX_SPACE_ID: u32 = 288;
const MAX_LOG_LEVEL: LevelFilter = LevelFilter::Trace;

#[derive(Debug, Deserialize, Serialize, tlua::Push, tlua::PushInto)]
pub struct DMLDelete {
    pub space: String,
    pub key: Vec<u8>,
}

#[derive(Debug, Deserialize, Serialize, tlua::Push, tlua::PushInto)]
pub struct DMLReplace {
    pub space: String,
    pub tuple: Vec<u8>,
}

#[derive(Debug, Deserialize, Serialize, tlua::Push, tlua::PushInto)]
pub enum DDLType {
    Insert,
    Update,
    Delete,
    Replace,
}

#[derive(Debug, Deserialize, Serialize, tlua::Push, tlua::PushInto)]
pub struct DDL {
    pub space_id: u32,
    pub key: Vec<u8>,
    pub data: Vec<u8>,
    pub kind: DDLType,
    pub target_space_name: Option<String>,
}

#[derive(Debug, Deserialize, Serialize)]
struct SpaceTuple {
    pub space_id: u32,
    pub owner_id: u64,
    pub space_name: String,
    pub engine: String,
    pub field_count: u64,
    pub flags: collections::HashMap<String, rmpv::Value>,
    pub format: Vec<collections::HashMap<String, rmpv::Value>>,
}

#[derive(Debug, Deserialize, Serialize, tlua::Push, tlua::PushInto)]
struct Key(Vec<u8>);

#[derive(Debug, Deserialize, Serialize, tlua::Push, tlua::PushInto)]
struct Data(Vec<u8>);

impl tuple::ToTupleBuffer for Key {
    fn write_tuple_data(&self, w: &mut impl std::io::Write) -> tarantool::Result<()> {
        w.write_all(&self.0)?;
        Ok(())
    }
}

impl tuple::ToTupleBuffer for Data {
    fn write_tuple_data(&self, w: &mut impl std::io::Write) -> tarantool::Result<()> {
        w.write_all(&self.0)?;
        Ok(())
    }
}

impl tuple::ToTupleBuffer for DMLReplace {
    fn write_tuple_data(&self, w: &mut impl std::io::Write) -> tarantool::Result<()> {
        w.write_all(&self.tuple)?;
        Ok(())
    }
}
impl tuple::ToTupleBuffer for DMLDelete {
    fn write_tuple_data(&self, w: &mut impl std::io::Write) -> tarantool::Result<()> {
        w.write_all(&self.key)?;
        Ok(())
    }
}

#[must_use]
fn value_to_vec(value: &rmpv::Value) -> Vec<u8> {
    let mut result = Vec::new();
    rmpv::encode::write_value(&mut result, value).expect("encode value should not fail");
    result
}

#[must_use]
fn vec_to_value(v: Vec<u8>) -> rmpv::Value {
    let mut cursor = std::io::Cursor::new(v);
    rmpv::decode::read_value(&mut cursor).expect("decode value should not fail")
}

static LOGGER: TarantoolLogger = TarantoolLogger::new();

fn setup_logger() {
    LOGGER.enabled(&log::Metadata::builder().build());
    if log::set_logger(&LOGGER).is_ok() {
        log::set_max_level(MAX_LOG_LEVEL);
    } else {
        log::info!("Logger has already been initialized");
    }
}

#[proc]
fn init() {
    setup_logger();
    log::info!("Ouroboros Rust initialized");
}

#[proc]
fn apply_dml_replace(replace: DMLReplace) -> Result<bool, Box<dyn Error>> {
    if let Some(space) = space::Space::find(&replace.space) {
        space.replace(&replace)?;
        Ok(true)
    } else {
        log::error!("Apply REPLACE: Can not find {} space", &replace.space);
        Err(format!("Can not find {} space", &replace.space).into())
    }
}

#[proc]
fn apply_dml_delete(delete: DMLDelete) -> Result<bool, Box<dyn Error>> {
    if let Some(space) = space::Space::find(&delete.space) {
        space.delete(&delete)?;
        Ok(true)
    } else {
        log::error!("Apply DELETE: Can not find {} space", &delete.space);
        Err(format!("Can not find {} space", &delete.space).into())
    }
}

fn get_space_id(space_name: &str) -> Option<u32> {
    log::debug!("space to search: {space_name}");
    space::Space::find(space_name).map(|space| space.id())
}

fn generate_space_id(raw_tuple: &rmpv::Value) -> Result<u32, Box<dyn Error>> {
    let arr = raw_tuple.as_array().expect("tuple should be an array");
    let flags = arr
        .get(5)
        .expect("tuple should have at least 6 elements")
        .as_map()
        .expect("element 5 of tuple should be a map");
    let mut is_temporary = false;
    for (flag_name, flag_value) in flags {
        match flag_name.as_str() {
            Some("type") => {
                let flag_value = flag_value.as_str().unwrap_or_default();
                if flag_value == "temporary" {
                    is_temporary = true;
                }
            }
            _ => continue,
        }
    }
    tarantool::schema::space::generate_space_id(is_temporary)
        .map_err(|e| format!("can not generate space id: {e:?}").into())
}

#[proc]
fn apply_ddl(ddl: DDL) -> Result<bool, Box<dyn Error>> {
    let mut ddl = ddl;
    match ddl.kind {
        DDLType::Insert | DDLType::Update => {
            log::debug!("{:?}", ddl);
            unreachable!("should be REPLACE and DELETE only")
        }
        DDLType::Replace => {
            log::debug!(
                "space_id: {}, target_name: {:?}, data: {}",
                ddl.space_id,
                ddl.target_space_name,
                vec_to_value(ddl.data.clone()),
            );
            // If it is a space replace
            if ddl.space_id == SPACE_SPACE_ID {
                let raw_tuple = vec_to_value(ddl.data);
                let arr = raw_tuple.as_array().expect("tuple ashould be an array");
                let space_name = arr
                    .get(2)
                    .expect("tuple should have at least 3 elements")
                    .as_str()
                    .expect("element 3 of tuple should be a string");
                let space_id = if let Some(r) = get_space_id(space_name) {
                    r
                } else {
                    log::debug!("can not find space {space_name} in cache map");
                    generate_space_id(&raw_tuple)?
                };

                let mut new_arr = Vec::with_capacity(arr.len());
                new_arr.push(rmpv::Value::Integer(space_id.into()));
                for item in arr.iter().skip(1) {
                    new_arr.push(item.clone());
                }

                ddl.data = value_to_vec(&rmpv::Value::Array(new_arr));
            }
            // If it is a new index creation
            if ddl.space_id == INDEX_SPACE_ID {
                let raw_tuple = vec_to_value(ddl.data);
                let arr = raw_tuple.as_array().expect("tuple should be an array");
                let space_name = ddl
                    .target_space_name
                    .as_ref()
                    .expect("Index DDL should have non-empty target_space_name");
                if let Some(space_id) = get_space_id(space_name) {
                    let mut new_arr = Vec::with_capacity(arr.len());
                    new_arr.push(rmpv::Value::Integer(space_id.into()));
                    for item in arr.iter().skip(1) {
                        new_arr.push(item.clone());
                    }
                    ddl.data = value_to_vec(&rmpv::Value::Array(new_arr));
                } else {
                    return Err(format!("Unknown space_name {space_name}").into());
                }
            }

            let space = unsafe { space::Space::from_id_unchecked(ddl.space_id) };
            let res = space.replace(&Data(ddl.data));
            if res.is_ok() {
                return Ok(true);
            }
            match res.err() {
                Some(tarantool::error::Error::Tarantool(box_error)) => {
                    log::error!("Apply DDL REPLACE should not fail: {box_error:?}");
                    return Err(
                        format!("replacing ddl should not fail: tarantool: {box_error:?}").into(),
                    );
                }
                err => {
                    log::error!("Apply DDL REPLACE should not fail: {err:?}");
                    return Err(format!("replacing ddl should not fail: {err:?}").into());
                }
            };
        }
        DDLType::Delete => {
            let mut key = None;
            log::debug!(
                "space_id: {}, data: {}",
                ddl.space_id,
                vec_to_value(ddl.key.clone()),
            );

            // DELETE SPACE ID WITH ACTUAL SPACE ID IF IT EXISTS
            if ddl.space_id == SPACE_SPACE_ID || ddl.space_id == INDEX_SPACE_ID {
                let space_name = ddl
                    .target_space_name
                    .as_ref()
                    .expect("DDL DELETE should have target_name_space");
                if let Some(space_id) = get_space_id(space_name) {
                    let raw_tuple = vec_to_value(ddl.key.clone());
                    let arr = raw_tuple.as_array().expect("tuple should be an array");
                    let mut new_arr = Vec::with_capacity(arr.len());
                    new_arr.push(rmpv::Value::Integer(space_id.into()));
                    for item in arr.iter().skip(1) {
                        new_arr.push(item.clone());
                    }
                    key = Some(value_to_vec(&rmpv::Value::Array(new_arr)));
                }
            }

            if let Some(key_val) = key {
                let space = unsafe { space::Space::from_id_unchecked(ddl.space_id) };
                space.delete(&Key(key_val))?;
            }
        }
    };

    Ok(true)
}

#[proc]
fn test_lib() -> bool {
    true
}
