local json = require('json')
local fio = require('fio')
local fun = require('fun')
local httpc = require('http.client').new()

local DEFAULT_SETTINGS = {
    host = 'localhost',
    port = '8080',
    prefix = '/ouroboros/api/v1',
}

local common = {}

--- Parses HTTP request response body into lua-object
--- @param http_body any
--- @param http_content_type? string
--- @return table|nil @ lua-object
local function parse_response_body(http_body, http_content_type)
    http_content_type = http_content_type or 'application/json'

    if string.match(http_content_type, 'text') then
        return http_body
    end

    local ok, res = pcall(json.decode, http_body)
    return (ok and res) or nil
end

--- Generates url using module settings. It constists of host, port, endpoint prefix and path itself
--- @param self table
--- @param path string
--- @return string url
function common.url(self, path, absolute)
    path = absolute and path or fio.pathjoin(self.prefix, path)
    return string.format('http://%s:%s%s', self.host, self.port, path)
end

--- Perform HTTP request.
--- @param self table
--- @param method string
--- @param path string
--- @return table @ Response object from HTTP client.
function common.http_request(self, method, path, opts)
    opts = opts or {}
    local body = opts.body
    local absolute_path = path:startswith('/')
    local response = httpc:request(method, common.url(self, path, absolute_path), body)

    if response.body then
        response.json = parse_response_body(response.body, response.headers['content-type'])
    end

    return response
end

function common.metrics(self)
    return common.http_request(self, 'get', 'metrics')
end

function common.prometheus(self)
    return common.http_request(self, 'get', 'prometheus')
end

function common.health(self)
    return common.http_request(self, 'get', 'health')
end

local function new(opts)
    opts = opts or {}
    local settings = table.copy(DEFAULT_SETTINGS)
    local instance_settings = fun.chain(settings, opts):tomap()
    local instance_methods = fun.chain(common):tomap()
    local instance = setmetatable(instance_settings, {
        __index = instance_methods,
    })
    return instance
end

local function setup(opts)
    opts = opts or {}
    for k, v in pairs(opts) do
        DEFAULT_SETTINGS[k] = v
    end
end

local class_methods = fun.chain(common):tomap()

local Requests = setmetatable(DEFAULT_SETTINGS, {
    __index = class_methods
})

local function bind_state(fn, state)
    return function(...)
        return fn(state, ...)
    end
end

-- Wrap class methods with Requests state so all methods
-- (both class and instance) could use the same api
fun.iter(class_methods):each(function(name, fn)
    class_methods[name] = bind_state(fn, Requests)
end)

-- Do not need to wrap class methods with context
class_methods.new = new
class_methods.setup = setup

return Requests
